from .in_memory import InMemoryConnection
from .rabbit import RabbitMQConnection


async def connect(host="localhost", port=5672, broker="rabbit"):
    '''
    Create a connection object and then connect to a broker

    :param host: the hostname of the broker you wish to connect to
    :type host: str
    :param port: the port of the broker you wish to connect to
    :type port: int
    :param broker: broker type. Currently supported broker types are "rabbit"
        for a RabbitMQ broker and "in_memory" for an broker that resides in memory
        (useful for unit testing)
    :type broker: str
    :return: :class:`InMemoryConnection` or :class:`RabbitMQConnection` object

    .. todo:: raises BrokerConnectionError if cannot connect to the broker

    '''

    if broker == "in_memory":
        conn = InMemoryConnection(host=host, port=port)
    elif broker == "rabbit":
        conn = RabbitMQConnection(host=host, port=port)
    else:
        raise NotImplementedError

    await conn.connect()
    return conn



import subprocess
import time
import json
import os
from functools import wraps, partial
import asyncio

import pika

from . import base


class RabbitMQBroker(base.Broker):
    '''
    Control an existing RabbitMQBroker on your machine.

    Useful when performing integration testing of projects that depend on RabbitMQ.
    '''

    def _run(self):
        if os.environ["TEST_DISTRIBUTION"] == "arch":
            subprocess.run("sudo systemctl restart rabbitmq.service", shell=True, check=True, timeout=5)
        elif os.environ["TEST_DISTRIBUTION"] == "ubuntu":
            subprocess.run("sudo service rabbitmq-server restart", shell=True, check=True, timeout=5)
        else:
            raise ValueError("TEST_DISTRIBUTION environmental variable not set to either arch or ubuntu")

    async def run(self, is_running=None):
        '''
        Restarts the RabbitMQ broker using a method derived from the TEST_DISTRIBUTION
        environmental variable.

        :param is_running: A future set to done once the broker is confirmed
            as being running
        :type is_running: future

        If TEST_DISTRIBUTION=="arch", will try to restart rabbitmq using the linux ``systemctl``
        command.

        If TEST_DISTRIBUTION=="ubuntu", will try to restart rabbitmq using the linux ``service``
        command.

        Will wait for 20 seconds after restarting before returning.

        :raises ValueError: if TEST_DISTRIBUTION environmental variable not found

        .. note:: the user who invokes this method will likely require sudo access to the linux commands.
            This can be provided by editing the sudoers file.
        '''

        await self.loop.run_in_executor(None, self._run)
        await asyncio.sleep(20)
        if is_running:
            is_running.set_result(None)


class RabbitMQConnection(base.Connection):
    '''
    Implementation of a connection to a RabbitMQ broker.
    '''
    def __init__(self, **kwargs):
        '''
        :param host: the hostname of the broker you wish to connect to
        :type host: str
        :param port: the port of the broker you wish to connect to
        :type port: int

        .. note:: must call :meth:`RabbitMQConnection.connect` to actually connect to the broker
        '''

        super().__init__(**kwargs)
        self.channel_resource_constructor_func = RabbitMQChannel

    def _connect(self):
        cp = pika.ConnectionParameters(host=self.host, port=self.port)
        self._conn = pika.BlockingConnection(cp)

    async def connect(self):
        '''
        Connect to the RabbitMQ broker
        '''
        await self.loop.run_in_executor(None, self._connect)

    @base.lock_connection
    def _create_channel(self):
        internal_chan = self._conn.channel()

        chan = RabbitMQChannel(internal_chan=internal_chan, loop=self.loop)
        self.channels.append(chan)
        return chan

    async def create_channel(self):
        '''
        create a channel for multiplexing the connection

        :returns: a :class:`RabbitMQChannel` object
        '''
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, self._create_channel)

    def close(self):
        '''
        Stop processing events and close the connection to the broker

        :raises NotImplementedError:
        '''
        raise NotImplementedError

    def _process_events(self, num_cycles=None):
        while True:
            with self.conn_lock:
                self._conn.process_data_events(time_limit=0.1)

            if num_cycles is not None:
                num_cycles = num_cycles - 1
                if num_cycles == 0:
                    break

    async def process_events(self, num_cycles=None):
        '''
        Receive messages from the RabbitMQ broker and schedule
        associated callback couroutines.

        Should be run as a task in your app and not awaited for.

        :param num_cycles: the number of times to run the
            event processing loop. A value of None
            will cause events to be processed without a cycle limit.
        :type num_cycles: int|None
        '''
        loop = asyncio.get_event_loop()
        func = partial(self._process_events, num_cycles=num_cycles)
        await loop.run_in_executor(None, func)


class RabbitMQChannel(base.Channel):
    '''
    Implementation of a RabbitMQ Channel
    '''

    @base.lock_channel
    def _register_producer(self, *, exchange_name, exchange_type):
        self._chan.exchange_declare(exchange=exchange_name,
                                    type=exchange_type)

    async def register_producer(self, *, exchange_name, exchange_type):
        '''
        Register a producer on the channel by providing information to
        the broker about the exchange the channel is going to use.

        :param exchange_name: name of the exchange
        :type exchange_name: str
        :param exchange_type: Type of the exchange. Accepted values are "direct",
            "topic" or "fanout"
        :type exchange_type: str
        :returns: None
        '''

        loop = asyncio.get_event_loop()
        func = partial(self._register_producer, exchange_name=exchange_name,
                       exchange_type=exchange_type)

        await loop.run_in_executor(None, func)

    @base.lock_channel
    def _register_consumer(self, *, exchange_name, exchange_type, queue_name, callback, routing_keys=[""]):
        self._chan.exchange_declare(exchange=exchange_name,
                                    type=exchange_type)

        exclusive = False
        if queue_name is None:
            exclusive = True
            method_frame = self._chan.queue_declare(exclusive=True)
            queue_name = method_frame.method.queue
        else:
            self._chan.queue_declare(queue=queue_name)

        for r in routing_keys:
            self._chan.queue_bind(exchange=exchange_name,
                                  queue=queue_name,
                                  routing_key=r)

        pika_callback = self._wrap_callback(callback)
        self._chan.basic_consume(pika_callback,
                                 queue=queue_name,
                                 no_ack=True,
                                 exclusive=exclusive,
                                 consumer_tag=None,
                                 arguments=None)

    async def register_consumer(self, *, exchange_name, exchange_type, queue_name, callback, routing_keys=[""]):
        '''
        Register a consumer on the RabbitMQ channel.

        :param exchange_name: name of the exchange
        :param exchange_type: Type of the exchange. Accepted values are "direct",
            "topic" or "fanout"
        :param queue_name: name of the queue. If None, a name will be given
            automatically and the queue will be declared exclusive to the channel,
            meaning it will be deleted once the channel is closed.
        :param callback: The callback to run when a message is placed on the queue that
            matches one of the routing keys
        :param routing_keys: A list of keys to match against. A message will only be sent
            to a consumer if its routing key matches one or more of the routing keys listed

        :type exchange_name: str
        :type exchange_type: str
        :type queue_name: str|None
        :type callback: coroutine
        :type routing_keys: [str,]
        '''

        loop = asyncio.get_event_loop()
        func = partial(self._register_consumer, exchange_name=exchange_name,
                       exchange_type=exchange_type, queue_name=queue_name,
                       callback=callback, routing_keys=routing_keys)

        await loop.run_in_executor(None, func)

    def _wrap_callback(self, async_callback):
        '''
        Decorator to turn a pika callback running outside the
        main thread into a coroutine running in the main thread
        with simplified arguments.
        '''

        def thread_callback(ch, meth, prop, body):
            def main_loop_callback(ch, meth, prop, body):
                resp = {
                    "msg": json.loads(body),
                }

                return base.create_task(async_callback(resp), self.loop)

            return self.loop.call_soon_threadsafe(main_loop_callback, ch, meth, prop, body)

        return thread_callback

    @base.lock_channel
    def _publish(self, *, exchange_name, msg, routing_key=''):
        to_send_json = json.dumps(msg)
        self._chan.basic_publish(exchange=exchange_name,
                                 routing_key=routing_key,
                                 properties=None,
                                 body=to_send_json)

    async def publish(self, *, exchange_name, msg, routing_key=''):
        '''
        Publish a message on the channel.

        :param exchange_name: The name of the exchange to send the message to
        :param msg: The message to send
        :param routing_key: The routing key to associated the message with
        :type exchange_name: str
        :type msg: str
        :type routing_key: str
        '''
        loop = asyncio.get_event_loop()
        func = partial(self._publish, exchange_name=exchange_name,
                       msg=msg, routing_key=routing_key)

        await loop.run_in_executor(None, func)



'''
mooq is a asyncio compatible library for interacting with RabbitMQ AMQP broker.

'''

from .__version__ import __version__

from .base import ExchangeNotFound, ConsumerQueueNotFound, ConsumeTimeout, \
    NothingToConsume, BadExchange, BrokerInternalError

from .in_memory import InMemoryBroker, InMemoryConnection, InMemoryChannel

from .rabbit import RabbitMQBroker, RabbitMQConnection, RabbitMQChannel

from .connect import connect



# import unittest
import unittest
from unittest.mock import Mock, MagicMock
import os
import sys
import asyncio

import xmlrunner

from test import common
from test.unittest_utils import set_test_hang_alarm
from test.unittest_utils import clear_test_hang_alarm
from test.unittest_utils import close_all_threads
from test.unittest_utils import asyncio_test, AsyncMock

import mooq




# @unittest.skip("skipped") 
class InMemoryDirectProduceConsumeTest(common.TransportTestCase):
    async def async_setUp(self):
        await self.GIVEN_InMemoryBrokerStarted("localhost",1234)
        await self.GIVEN_ConnectionResourceCreated("localhost",1234,"in_memory")
        await self.GIVEN_ChannelResourceCreated()
        await self.GIVEN_ProducerRegistered(exchange_name="fake_exch",
                                            exchange_type="direct")


    async def async_tearDown(self):
        await self.CloseBroker()

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_callback_is_run(self):
        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue",
                                      exchange_name="fake_exch",
                                      exchange_type="direct",
                                      routing_keys=["fake_routing_key"],
                                      callback = self.fake_callback)
        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                    msg="fake_message",
                                    routing_key="fake_routing_key")
        await self.WHEN_ProcessEventsNTimes(2)

        self.THEN_CallbackReceivesMessage("fake_message")



    # @unittest.skip("skipped")
    @asyncio_test
    async def test_callback_is_not_run_if_routing_key_mismatch(self):
        fake_callback = Mock()
        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue",
                                      exchange_name="fake_exch",
                                      exchange_type="direct",
                                      routing_keys=["fake_routing_key"],
                                      callback = fake_callback)

        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                    msg="fake_message",
                                    routing_key="another_routing_key")

        await self.WHEN_ProcessEventsOnce()

        self.THEN_CallbackIsNotRun(fake_callback)


    # @unittest.skip("skipped")
    @asyncio_test
    async def test_callback_run_if_multiple_routing_keys(self):
        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue",
                                      exchange_name="fake_exch",
                                      exchange_type="direct",
                                      routing_keys=["fake_routing_key","fake_routing_key2"],
                                      callback = self.fake_callback)

        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                        msg="fake_message",
                                        routing_key="fake_routing_key2")

        await self.WHEN_ProcessEventsOnce()

        self.THEN_CallbackReceivesMessage("fake_message")


    # @unittest.skip("skipped")
    @asyncio_test
    async def test_callback_run_if_two_exclusive_queues_registered(self):
        await self.GIVEN_ConsumerRegistered(queue_name = None,
                                      exchange_name="fake_exch",
                                      exchange_type="direct",
                                      routing_keys=["fake_routing_key"],
                                      callback = self.fake_callback)

        await self.GIVEN_ConsumerRegistered(queue_name = None,
                                      exchange_name="fake_exch",
                                      exchange_type="direct",
                                      routing_keys=["fake_routing_key"],
                                      callback = self.fake_callback2)

        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                    msg="fake_message",
                                    routing_key="fake_routing_key")

        await self.WHEN_ProcessEventsOnce()

        self.THEN_CallbackReceivesMessage("fake_message")
        self.THEN_Callback2ReceivesMessage("fake_message")


    # @unittest.skip("skipped")
    @asyncio_test
    async def test_bad_exchange(self):
        fake_callback = Mock()

        with self.assertRaises(mooq.BadExchange):
            await self.WHEN_ConsumerRegistered( queue_name="fake_consumer_queue",
                                          exchange_name="fake_exch",
                                          exchange_type="fanout",
                                          routing_keys=["fake_routing_key"],
                                          callback = fake_callback)




# @unittest.skip("skipped") 
class DirectBadExchangeTest(common.TransportTestCase):
    async def async_setUp(self):
        await self.GIVEN_InMemoryBrokerStarted("localhost",1234)
        await self.GIVEN_ConnectionResourceCreated("localhost",1234,"in_memory")
        await self.GIVEN_ChannelResourceCreated()
        await self.GIVEN_ProducerRegistered(exchange_name="fake_exch",
                                      exchange_type="direct")


    async def async_tearDown(self):
        await self.CloseBroker()

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_bad_exchange(self):
        with self.assertRaises(mooq.BadExchange):
            await self.WHEN_ProducerRegistered(exchange_name="fake_exch",
                                               exchange_type="fanout")


class TopicTestCase(common.TransportTestCase):

    async def async_setUp(self):
        await self.GIVEN_InMemoryBrokerStarted("localhost",1234)
        await self.GIVEN_ConnectionResourceCreated("localhost",1234,"in_memory")
        await self.GIVEN_ChannelResourceCreated()
        await self.GIVEN_ProducerRegistered(exchange_name="fake_exch",
                                      exchange_type="topic")

    async def async_tearDown(self):
        await self.CloseBroker()




# @unittest.skip("skipped") 
class InMemoryTopicBallColourTest(TopicTestCase):
    async def async_setUp(self):
        await super().async_setUp()
        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue",
                                      exchange_name="fake_exch",
                                      exchange_type="topic",
                                      routing_keys=["ball.*","*.red"],
                                      callback = self.fake_callback)
    async def async_tearDown(self):
        await super().async_tearDown()

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_ball_yellow(self):
        await self.GWT_BallColour_RunsCallback("ball.yellow")

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_ball_red(self):
        await self.GWT_BallColour_RunsCallback("ball.red")

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_apple_red(self):
        await self.GWT_BallColour_RunsCallback("apple.red")

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_apple_yellow(self):
        await self.GWT_BallColour_DoesntRunCallback("apple.yellow")

    async def GWT_BallColour(self,*,routing_key,callback_run):

        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                    msg="fake_msg",
                                    routing_key=routing_key)

        await self.WHEN_ProcessEventsOnce()

        if callback_run:
            self.THEN_CallbackReceivesMessage("fake_msg")
        else:
            self.THEN_CallbackDoesntReceiveMessage()

    async def GWT_BallColour_RunsCallback(self,routing_key):
        await self.GWT_BallColour(routing_key=routing_key,callback_run=True)

    async def GWT_BallColour_DoesntRunCallback(self,routing_key):
        await self.GWT_BallColour(routing_key=routing_key,callback_run=False)


# @unittest.skip("skipped") 
class InMemoryTopicAllWildcardsTest(TopicTestCase):
    async def async_setUp(self):
        await super().async_setUp()
        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue",
                                      exchange_name="fake_exch",
                                      exchange_type="topic",
                                      routing_keys=["*.*"],
                                      callback = self.fake_callback)

    async def async_tearDown(self):
        await super().async_tearDown()

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_runs_callback(self):
        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                    msg="fake_msg",
                                    routing_key="absolutely.everything")

        await self.WHEN_ProcessEventsOnce()

        self.THEN_CallbackReceivesMessage("fake_msg")


# @unittest.skip("skipped") 
class InMemoryTopicRunTwiceTest(TopicTestCase):
    async def async_setUp(self):
        await super().async_setUp()
        self.fake_callback = AsyncMock()
        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue",
                                      exchange_name="fake_exch",
                                      exchange_type="topic",
                                      routing_keys=["*.*","ball.red"],
                                      callback = self.fake_callback)


    # @unittest.skip("skipped")
    @asyncio_test
    async def test_runs_callback_twice(self):
        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                    msg="fake_msg",
                                    routing_key="ball.red")

        await self.WHEN_ProcessEventsNTimes(3)

        self.THEN_CallbackIsRun(self.fake_callback,num_times=2)

# @unittest.skip("skipped") 
class FanoutTestCase(common.TransportTestCase):
    async def async_setUp(self):
        await super().async_setUp()
        await self.GIVEN_InMemoryBrokerStarted("localhost",1234)
        await self.GIVEN_ConnectionResourceCreated("localhost",1234,"in_memory")
        await self.GIVEN_ChannelResourceCreated()
        await self.GIVEN_ProducerRegistered(exchange_name="fake_exch",
                                      exchange_type="fanout")


    async def async_tearDown(self):
        await self.CloseBroker()

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_routes_to_all_consumer_queues(self):
        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue1",
                                      exchange_name="fake_exch",
                                      exchange_type="fanout",
                                      routing_keys="",
                                      callback = self.fake_callback)

        await self.GIVEN_ConsumerRegistered(queue_name="fake_consumer_queue2",
                                      exchange_name="fake_exch",
                                      exchange_type="fanout",
                                      routing_keys="",
                                      callback = self.fake_callback2)

        await self.GIVEN_MessagePublished(exchange_name="fake_exch",
                                    msg="fake_msg",
                                    routing_key="")

        await self.WHEN_ProcessEventsNTimes(1)

        self.THEN_CallbackReceivesMessage("fake_msg")
        self.THEN_Callback2ReceivesMessage("fake_msg")


# @unittest.skip("skipped") 
class ExchangeDoesntExistTest(common.TransportTestCase):
    async def async_setUp(self):
        await super().async_setUp()
        await self.GIVEN_InMemoryBrokerStarted("localhost",1234)
        await self.GIVEN_ConnectionResourceCreated("localhost",1234,"in_memory")
        await self.GIVEN_ChannelResourceCreated()

    async def async_tearDown(self):
        await self.CloseBroker()

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_publish_to_exchange_that_doesnt_exist(self):
        with self.assertRaises(mooq.BadExchange):
            await self.WHEN_MessagePublished(exchange_name="fake_exch",
                                        msg="fake_message",
                                        routing_key="fake_routing_key")











if __name__ == '__main__':
    unittest.main(
        testRunner=xmlrunner.XMLTestRunner(output='test-reports'),
        # these make sure that some options that are not applicable
        # remain hidden from the help menu.
        failfast=False, buffer=False, catchbreak=False)


from collections import namedtuple
import queue
import uuid
import traceback
import re
import asyncio
from . import base


InMemoryChannelInternal = namedtuple("InMemoryChannelInternal", ["msg_q", "broker_q"])


class InMemoryBroker(base.Broker):
    '''
    Implementation of an in memory broker
    '''

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.broker_ctl_q = queue.Queue()
        self.msg_q = queue.Queue()
        self.broker_q = queue.Queue()
        self.exchanges = []
        self.consumer_queues = []

    async def close(self):
        '''
        close the broker
        '''

        self.broker_ctl_q.put("close", block=False)
        await self.not_running

    async def run(self, is_running=None):
        '''
        restarts the broker

        :param is_running: A future set to done once the broker is confirmed
            as being running
        :type is_running: future
        '''

        if not is_running:
            raise ValueError("is running future must be supplied as keyword argument")

        is_running.set_result(None)
        self.not_running = self.loop.create_future()
        while True:
            try:
                broker_ctl_cmd = self.broker_ctl_q.get(block=False)
            except queue.Empty:
                pass
            else:
                if broker_ctl_cmd == "close":
                    self.not_running.set_result(None)
                    break

            try:
                msg = self.msg_q.get(block=False)
            except queue.Empty:
                pass
            else:
                self._handle_msg(msg)

            await asyncio.sleep(0.001)

    def _handle_msg(self, msg):
        try:
            self._command_router(msg)
        except Exception as e:
            tb_str = traceback.format_exc()
            self._put_error(base.BrokerInternalError, msg=tb_str)

    def _command_router(self, msg):
        if msg['command'] == "publish":
            self._route_message_to_consumer_queues(
                msg['exchange_name'],
                msg['msg'],
                msg['routing_key']
            )

        elif msg['command'] == "register_producer":
            self._add_exchange(msg['exchange_name'], msg['exchange_type'])
        elif msg['command'] == "register_consumer":
            self._add_exchange(msg['exchange_name'], msg['exchange_type'])
            self._add_consumer_queue(msg['queue_name'])
            exch = self._get_exchange(msg['exchange_name'])
            if msg['exchange_type'] == "fanout":
                msg['routing_keys'] = [""]

            exch.bind(msg['queue_name'], msg['routing_keys'])

    def _add_exchange(self, name, type_):
        try:
            exch = self._get_exchange(name)
        except base.ExchangeNotFound:
            exch = _InMemoryExchange(name=name, type_=type_)
            self.exchanges.append(exch)
        else:
            if type_ != exch.type_:
                msg = ("tried to declare an exchange of type {}"
                     " when it has already been declared of typ {}"
                     "").format(type_, exch.type_)

                self._put_error(base.BadExchange, msg=msg)

    def _put_error(self, e, msg=""):
        self.broker_q.put({"response": "error",
                           "error": e,
                           "msg": msg}, block=False)

    def _get_exchange(self, name):
        for exch in self.exchanges:
            if exch.name == name:
                return exch

        raise base.ExchangeNotFound

    def _add_consumer_queue(self, name):
        if not self._consumer_queue_exists(name):
            cq = _InMemoryConsumerQueue(name=name)
            self.consumer_queues.append(cq)

    def _get_consumer_queue(self, name):
        for cq in self.consumer_queues:
            if cq.name == name:
                return cq

        raise base.ConsumerQueueNotFound

    def _consumer_queue_exists(self, name):
        try:
            self._get_consumer_queue(name)
            return True
        except base.ConsumerQueueNotFound:
            return False

    def _route_message_to_consumer_queues(self, exchange_name, msg, routing_key):
        try:
            exch = self._get_exchange(exchange_name)
        except base.ExchangeNotFound:
            self._put_error(base.BadExchange)
        else:
            if exch.type_ == "direct":
                self._route_direct(exch, msg, routing_key)
            elif exch.type_ == "topic":
                self._route_topic(exch, msg, routing_key)
            elif exch.type_ == "fanout":
                self._route_fanout(exch, msg)
            else:
                raise NotImplementedError

    def _route_direct(self, exch, msg, routing_key):
        for queue_name, routing_keys in exch.queues.items():
            if routing_key in routing_keys:
                cq = self._get_consumer_queue(queue_name)
                cq.put({"msg": msg, "routing_key": routing_key})

    def _route_topic(self, exch, msg, routing_key):
        for queue_name, routing_keys in exch.queues.items():
            for r in routing_keys:
                pattern = self._regexify(r)
                if pattern.match(routing_key):
                    cq = self._get_consumer_queue(queue_name)
                    cq.put({"msg": msg, "routing_key": routing_key})

    def _route_fanout(self, exch, msg):
        for queue_name in exch.queues:
            cq = self._get_consumer_queue(queue_name)
            cq.put({"msg": msg, "routing_key": ""})

    def _regexify(self, routing_key):
        splitted = routing_key.split(".")
        regex_elements = []
        for k in splitted:
            if k == "*":
                regex_elements.append(".+")
            else:
                regex_elements.append(k)

        regex_str = "\.".join(regex_elements)

        return re.compile(regex_str)


class _InMemoryExchange(base.Exchange):
    '''
    Implementation of an in memory exchange
    '''
    def __init__(self, **kwargs):
        '''
        :param name: name of exchange
        :param type\_: type of exchange. Accepted values are
            "direct", "fanout" or "topic"
        :type name: str
        :type type\_: str
        '''

        super().__init__(**kwargs)
        self.queues = {}

    def bind(self, queue_name, routing_keys):
        '''
        Bind routing keys to a queue

        :param queue_name: The name of the queue to bind
        :param routing_keys: A list of keys to match against to bind to the queue
        :type queue_name: str
        :type routing_keys: [str,]
        '''

        if queue_name not in self.queues:
            self.queues[queue_name] = routing_keys
        else:
            self.queues[queue_name].extend(routing_keys)


class _InMemoryConsumerQueue(base.ConsumerQueue):
    '''
    Implementation of an in memory consumer queue
    '''
    def __init__(self, **kwargs):
        '''
        :param name: the name of the queue
        :type name: str
        '''

        super().__init__(**kwargs)
        self._q = queue.Queue()
        self.callback = None

    def get_next_message(self):
        '''
        Get the next message from the consumer queue.

        :return: a message string

        :raises: :class:`ConsumeTimeout` if no messages on the queue
        '''

        try:
            msg_dict = self._q.get(block=False)
            return msg_dict
        except queue.Empty:
            raise base.ConsumeTimeout

    def put(self, data):
        '''
        Put data directly into the consumer queue

        :param data: data to add
        :type data: str
        '''

        self._q.put(data, block=False)


class InMemoryConnection(base.Connection):
    '''
    Implementation of an in memory connection to a broker
    '''

    def __init__(self, **kwargs):
        '''
        :param host: the hostname of the broker you wish to connect to
        :type host: str
        :param port: the port of the broker you wish to connect to
        :type port: int

        .. note:: must call :meth:`InMemoryConnection.connect` to actually connect to the broker
        '''

        super().__init__(**kwargs)

    async def connect(self):
        '''
        Connect to the InMemory broker
        '''
        self.broker = self.get_broker(host=self.host, port=self.port)
        self.msg_q = self.broker.msg_q
        self.broker_q = self.broker.broker_q

    # @base.lock_connection
    async def create_channel(self):
        '''
        create a channel for multiplexing the connection
        :returns: an :class:`InMemoryChannel` object
        '''
        internal_chan = InMemoryChannelInternal(msg_q=self.msg_q, broker_q=self.broker_q)

        chan = InMemoryChannel(internal_chan=internal_chan, loop=self.loop)
        self.channels.append(chan)
        return chan

    async def close(self):
        '''
        Stop processing events and close the connection to the broker
        '''
        self.connected = False

    async def process_events(self, num_cycles=None):
        '''
        Receive messages from the broker and schedule
        associated callback couroutines.

        :param num_cycles: the number of times to run the
            event processing loop. A value of None
            will cause events to be processed without a cycle limit.
        :type num_cycles: int|None
        '''

        while True:
            for chan in self.channels:
                for queue_name, callback in chan.callbacks.items():
                    cq = self.broker._get_consumer_queue(queue_name)

                    try:
                        msg_dict = cq.get_next_message()
                    except base.ConsumeTimeout:
                        pass
                    else:
                        _, launched = base.create_task(callback(msg_dict), self.loop)
                        await launched

            if num_cycles is not None:
                num_cycles = num_cycles - 1
                if num_cycles == 0:
                    break


class InMemoryChannel(base.Channel):
    '''
    Implementation of an in memory broker channel
    '''

    def __init__(self, **kwargs):
        '''
        :param internal_chan: the transport specific channel object to use
        :param loop: the event loop

        Typically this class will be instantiated outside the main thread.
        '''
        super().__init__(**kwargs)
        self.callbacks = {}

    # @base.lock_channel
    async def register_producer(self, *, exchange_name, exchange_type):
        '''
        Register a producer on the channel by providing information to
        the broker about the exchange the channel is going to use.

        :param exchange_name: name of the exchange
        :type exchange_name: str
        :param exchange_type: Type of the exchange. Accepted values are "direct",
            "topic" or "fanout"
        :type exchange_type: str
        '''
        self._chan.msg_q.put(
            {
                "command": "register_producer",
                "exchange_name": exchange_name,
                "exchange_type": exchange_type,
            }, block=False
        )
        await self._handle_broker_responses()

    # @base.lock_channel
    async def register_consumer(self, *, exchange_name, exchange_type,
                         queue_name, callback, routing_keys=[""]):
        '''
        Register a consumer on the channel.

        :param exchange_name: name of the exchange
        :param exchange_type: Type of the exchange. Accepted values are "direct",
            "topic" or "fanout"
        :param queue_name: name of the queue. If None, a name will be given
            automatically and the queue will be declared exclusive to the channel,
            meaning it will be deleted once the channel is closed.
        :param callback: The callback to run when a message is placed on the queue that
            matches one of the routing keys
        :param routing_keys: A list of keys to match against. A message will only be sent
            to a consumer if its routing key matches one or more of the routing keys listed

        :type exchange_name: str
        :type exchange_type: str
        :type queue_name: str|None
        :type callback: coroutine
        :type routing_keys: [str,]

        '''

        if queue_name is None:
            queue_name = str(uuid.uuid4())

        self._chan.msg_q.put(
            {
                "command": "register_consumer",
                "exchange_name": exchange_name,
                "exchange_type": exchange_type,
                "queue_name": queue_name,
                "routing_keys": routing_keys
            }, block=False
        )

        self.callbacks[queue_name] = callback

        await self._handle_broker_responses()

    async def _handle_broker_responses(self):
        await asyncio.sleep(0.01)
        while True:
            try:
                resp = self._chan.broker_q.get(block=False)
            except queue.Empty:
                break
            else:
                if resp['response'] == "error":
                    raise resp['error'](resp['msg'])

            await asyncio.sleep(0.001)

    # @base.lock_channel
    async def publish(self, *, exchange_name, msg, routing_key=''):
        '''
        Publish a message on the channel.

        :param exchange_name: The name of the exchange to send the message to
        :param msg: The message to send
        :param routing_key: The routing key to associated the message with
        :type exchange_name: str
        :type msg: str
        :type routing_key: str
        '''
        self._chan.msg_q.put(
            {
                "command": "publish",
                "exchange_name": exchange_name,
                "msg": msg,
                "routing_key": routing_key,
            }, block=False
        )
        await self._handle_broker_responses()



from setuptools import setup
import os

here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, 'mooq', '__version__.py'), 'r') as f:
    exec(f.read(), about)

setup(
    name='mooq',
    version=about['__version__'],
    packages=['mooq'],
    description="An asyncio compatible library for interacting with a RabbitMQ AMQP broker",
    long_description="An asyncio compatible library for interacting with a RabbitMQ AMQP broker",
    include_package_data=True,
    install_requires=[
        'pika',
    ],
    zip_safe=False,
    author="Jeremy Arr",
    author_email="jeremyarr@gmail.com",
    license="MIT",
    keywords=["amqp",
              "pika",
              "rabbitmq",
             ],
    url="https://github.com/jeremyarr/mooq",
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'Topic :: Communications',
    ]
)




import unittest
import os
import json
import sys
import asyncio

import xmlrunner

import mooq
from test import common
from test.unittest_utils import asyncio_test

# @unittest.skip("skipped")
class RabbitMQDirectProduceConsumeTest(common.TransportTestCase):
    @classmethod
    def setUpClass(cls):
        cls.GIVEN_RabbitMQBrokerStarted("localhost",5672)

    async def async_setUp(self):
        await super().async_setUp()
        await self.GIVEN_ConnectionResourceCreated("localhost",5672,"rabbit")
        await self.GIVEN_ChannelResourceCreated()


    # @unittest.skip("skipped")
    @asyncio_test
    async def test_direct_runs_callback(self):
        await self.GIVEN_ProducerRegistered(exchange_name="fake_direct_exch",
                                      exchange_type="direct")

        await self.GIVEN_ConsumerRegistered(queue_name="fake_direct_consumer_queue",
                                      exchange_name="fake_direct_exch",
                                      exchange_type="direct",
                                      routing_keys=["fake_routing_key"],
                                      callback = self.fake_callback)

        await self.GIVEN_MessagePublished(exchange_name="fake_direct_exch",
                                    msg="fake_message",
                                    routing_key="fake_routing_key")

        await self.WHEN_ProcessEventsOnce()

        self.THEN_CallbackReceivesMessage("fake_message")

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_topic_runs_callback(self):
        await self.GIVEN_ProducerRegistered(exchange_name="fake_topic_exch",
                                      exchange_type="topic")

        await self.GIVEN_ConsumerRegistered(queue_name="fake_topic_consumer_queue",
                                      exchange_name="fake_topic_exch",
                                      exchange_type="topic",
                                      routing_keys=["*.red","ball.*"],
                                      callback = self.fake_callback)

        await self.GIVEN_MessagePublished(exchange_name="fake_topic_exch",
                                    msg="fake_message",
                                    routing_key="apple.red")

        await self.WHEN_ProcessEventsOnce()

        self.THEN_CallbackReceivesMessage("fake_message")

    # @unittest.skip("skipped")
    @asyncio_test
    async def test_fanout_runs_callback(self):
        await self.GIVEN_ProducerRegistered(exchange_name="fake_fanout_exch",
                                      exchange_type="fanout")

        await self.GIVEN_ConsumerRegistered(queue_name="fake_fanout_consumer_queue1",
                                      exchange_name="fake_fanout_exch",
                                      exchange_type="fanout",
                                      routing_keys=[""],
                                      callback = self.fake_callback)

        await self.GIVEN_ConsumerRegistered(queue_name="fake_fanout_consumer_queue2",
                                      exchange_name="fake_fanout_exch",
                                      exchange_type="fanout",
                                      routing_keys=[""],
                                      callback = self.fake_callback2)

        await self.GIVEN_MessagePublished(exchange_name="fake_fanout_exch",
                                    msg="fake_message",
                                    routing_key="")

        await self.WHEN_ProcessEventsOnce()

        self.THEN_CallbackReceivesMessage("fake_message")
        self.THEN_Callback2ReceivesMessage("fake_message")




if __name__ == '__main__':
    unittest.main(
        testRunner=xmlrunner.XMLTestRunner(output='test-reports'),
        # these make sure that some options that are not applicable
        # remain hidden from the help menu.
        failfast=False, buffer=False, catchbreak=False)


'''
Base classes, helper functions and custom exceptions for mooq

'''

import threading
import asyncio
from functools import wraps


class ExchangeNotFound(Exception):
    pass


class ConsumerQueueNotFound(Exception):
    pass


class ConsumeTimeout(Exception):
    pass


class NothingToConsume(Exception):
    pass


class BadExchange(Exception):
    pass


class BrokerInternalError(Exception):
    pass


broker_registry = {}


def create_task(coro_obj, loop):
    '''
    wrapper for creating a task that can be used for waiting
    until a task has started.

    :param coro_obj: coroutine object to schedule
    :param loop: event loop
    :returns: a two element tuple where the first element
        is the task object. Awaiting on this will return when
        the coroutine object is done executing. The second element
        is a future that becomes done when the coroutine object is started.

    .. note:: must only be called from within the thread
        where the event loop resides
    '''
    async def task_wrapper(coro_obj, launched):
        launched.set_result(True)
        await coro_obj

    launched = loop.create_future()
    return loop.create_task(task_wrapper(coro_obj, launched)), launched


class Broker(object):
    '''
    Base class for a broker. Not to be used directly.
    '''

    def __init__(self, *, host, port):
        '''
        Add the broker to the registry. Each broker is given
        a unique name of "host_port" in the registry.

        :param host: the hostname of the broker you wish to connect to
        :type host: str
        :param port: the port of the broker you wish to connect to
        :type port: int
        '''

        self.host = host
        self.port = port
        self.name = "{}_{}".format(self.host, self.port)
        broker_registry[self.name] = self
        self.loop = asyncio.get_event_loop()

    async def close(self):
        '''
        close the broker

        :raises: NotImplementedError
        '''
        raise NotImplementedError

    async def run(self, is_running=None):
        '''
        restarts the broker

        :param is_running: A future set to done once the broker is confirmed
            as being running
        :type is_running: future

        :raises: NotImplementedError
        '''
        raise NotImplementedError


class Exchange(object):
    '''
    Base class for an exchange. Not to be used directly.
    '''

    def __init__(self, *, name, type_):
        '''
        :param name: name of exchange
        :param type\_: type of exchange. Accepted values are
            "direct", "fanout" or "topic"
        :type name: str
        :type type\_: str
        '''
        self.name = name
        assert type_ in ["direct", "fanout", "topic"]
        self.type_ = type_

    def bind(self, queue_name, routing_keys):
        '''
        Bind routing keys to a queue

        :param queue_name: The name of the queue to bind
        :param routing_keys: A list of keys to match against to bind to the queue
        :type queue_name: str
        :type routing_keys: [str,]
        :raises: NotImplementedError
        '''
        raise NotImplementedError


class ConsumerQueue(object):
    '''
    Base class for a consumer queue. Not to be used directly.
    '''
    def __init__(self, *, name):
        '''
        :param name: the name of the queue
        :type name: str
        '''

        self.name = name

    def get_next_message(self):
        '''
        Get the next message from the consumer queue.

        :return: a message string

        :raises: :class:`ConsumeTimeout` if no messages on the queue
        '''
        raise NotImplementedError

    def put(self, data):
        '''
        Put data directly into the consumer queue

        :param data: data to add
        :type data: str
        '''
        raise NotImplementedError


class Connection(object):
    '''
    Base class for a connection to a broker. Not to be used directly.
    '''
    def __init__(self, *, host, port):
        '''
        :param host: the hostname of the broker you wish to connect to
        :type host: str
        :param port: the port of the broker you wish to connect to
        :type port: int

        .. note:: must call :meth:`connect` to actually connect to the broker
        '''
        self.host = host
        self.port = port
        self.conn_lock = _TimeoutRLock(1)
        self.connected = False
        self.channels = []
        self.loop = asyncio.get_event_loop()

    async def create_channel(self):
        '''
        create a channel for multiplexing the connection

        :raises NotImplementedError:
        '''
        raise NotImplementedError

    async def connect(self):
        '''
        Connect to the broker

        :raises NotImplementedError:
        '''
        raise NotImplementedError

    async def close(self):
        '''
        Stop processing events and close the connection to the broker

        :raises NotImplementedError:
        '''

        raise NotImplementedError

    async def process_events(self, num_cycles=None):
        '''
        Receive messages from the broker and schedule
        associated callback couroutines.

        :param num_cycles: the number of times to run the
            event processing loop. A value of None or "inf"
            will cause events to be processed without a cycle limit.
        :type num_cycles: int|None|"inf"
        :raises NotImplementedError:
        '''
        raise NotImplementedError

    def get_broker(self, *, host, port):
        '''
        Get the :class:`Broker` object associated with the connection.

        :param host: the hostname of the broker
        :type host: str
        :param port: the port of the broker
        :type port: int

        :return: A :class:`Broker` object
        '''

        broker_name = "{}_{}".format(host, port)
        return broker_registry[broker_name]


class Channel(object):
    '''
    Base class for a channel of a connection. Not to be used directly.
    '''
    def __init__(self, *, internal_chan, loop):
        '''
        :param internal_chan: the transport specific channel object to use
        :param loop: the event loop

        Typically this class will be instantiated outside the main thread.
        '''
        self._chan = internal_chan
        self.chan_lock = _TimeoutRLock(1)
        self.loop = loop

    async def register_producer(self, *, exchange_name, exchange_type):
        '''
        Register a producer on the channel by providing information to
        the broker about the exchange the channel is going to use.

        :param exchange_name: name of the exchange
        :type exchange_name: str
        :param exchange_type: Type of the exchange. Accepted values are "direct",
            "topic" or "fanout"
        :type exchange_type: str
        '''
        raise NotImplementedError

    async def register_consumer(self, *, exchange_name, exchange_type,
                                queue_name, callback, routing_keys):

        '''
        Register a consumer on the channel.

        :param exchange_name: name of the exchange
        :param exchange_type: Type of the exchange. Accepted values are "direct",
            "topic" or "fanout"
        :param queue_name: name of the queue. If None, a name will be given
            automatically and the queue will be declared exclusive to the channel,
            meaning it will be deleted once the channel is closed.
        :param callback: The callback to run when a message is placed on the queue that
            matches one of the routing keys
        :param routing_keys: A list of keys to match against. A message will only be sent
            to a consumer if its routing key matches one or more of the routing keys listed

        :type exchange_name: str
        :type exchange_type: str
        :type queue_name: str|None
        :type callback: coroutine
        :type routing_keys: [str,]

        '''
        raise NotImplementedError

    async def publish(self, *, exchange_name, msg, routing_key=''):
        '''
        Publish a message on the channel.

        :param exchange_name: The name of the exchange to send the message to
        :param msg: The message to send
        :param routing_key: The routing key to associated the message with
        :type exchange_name: str
        :type msg: str
        :type routing_key: str
        '''
        raise NotImplementedError


class _TimeoutRLock(object):
    '''
    Context manager for a reentrant Lock with timeout
    '''
    def __init__(self, timeout):
        self.rlock_obj = threading.RLock()
        self.timeout = timeout

    def __enter__(self):
        self.rlock_obj.acquire(timeout=self.timeout)

    def __exit__(self, exc_type, exc_value, tb):
        self.rlock_obj.release()


def lock_channel(func):
    '''
    Deocorator to acquire the channel lock and
    release it after use.
    '''
    @wraps(func)
    def inner(self, *args, **kwargs):
        with self.chan_lock:
            out = func(self, *args, **kwargs)

            return out

    return inner


def lock_connection(func):
    '''
    Deocorator to acquire the connection lock and
    release it after use.
    '''
    @wraps(func)
    def inner(self, *args, **kwargs):
        with self.conn_lock:
            out = func(self, *args, **kwargs)

            return out

    return inner
